import 'package:async/async.dart';
import 'package:bloc/bloc.dart';
import 'package:connectivity/connectivity.dart';
import 'package:dio/dio.dart';
import 'package:meta/meta.dart';

import '../../knt_bloc.dart';
import '../defined.dart';
import '../model/exceptions.dart';
import '../utils/logger.dart';

part 'events.dart';
part 'states.dart';

/// Base handles normal events
abstract class BaseBloc extends Bloc<BaseEvent, BaseState> {
  final bool needLog;
  BaseEvent? _currentEvent;
  CancelableOperation? _cancelableOperation;

  BaseBloc({this.needLog = false}) : super(BaseInitState());

  @override
  Future<void> close() async {
    await _cancelableOperation?.cancel();
    return super.close();
  }

  @override
  void onTransition(Transition<BaseEvent, BaseState> transition) {
    super.onTransition(transition);
    if (needLog) {
      logger.d('🚒🚒🚒 TRANSITION BLOC [-$runtimeType-] 🚒🚒🚒\n$transition');
    }
  }

  @override
  Stream<BaseState> mapEventToState(BaseEvent event) async* {
    _currentEvent = event;
    if (event is BaseErrorEvent) {
      yield* handleBaseErrorEvent(event);
      return;
    }
    if (event is ResetBlocEvent) {
      yield* handleResetBlocEvent(event);
      return;
    }
    if (event is CancelRunningTaskEvent) {
      yield* handleCancelRunningTaskEvent(event);
      return;
    }
    yield* handleEvent(event).handleError(
      (exception) => handleError(
        exception,
        tag: event.tag,
      ),
    );
  }

  /// Throws error state
  Stream<BaseState> handleBaseErrorEvent(BaseErrorEvent event) async* {
    yield BaseErrorState(
      tag: event.tag,
      exception: event.exception,
    );
  }

  /// Needs to reset data bloc, includes cancelling all running async tasks
  @mustCallSuper
  Stream<BaseState> handleResetBlocEvent(ResetBlocEvent event) async* {
    await _cancelableOperation?.cancel();
  }

  /// Cancels running tasks
  Stream<BaseState> handleCancelRunningTaskEvent(
      CancelRunningTaskEvent event) async* {
    await _cancelableOperation?.cancel();
  }

  /// Handle incoming events, then map into corresponding state
  @mustCallSuper
  Stream<BaseState> handleEvent(BaseEvent event) async* {
    if (event.needCheckConnection) {
      await checkNetworkConnection();
    }
  }

  /// Handle exception which is thrown from mapping event-state process
  @mustCallSuper
  void handleError(exception, {String? tag}) {
    if (exception is KntException) {
      if (needLog) {
        logger.e('❌❌ KntException: $exception ❌❌');
      }
      add(BaseErrorEvent(
        tag: tag,
        exception: exception,
      ));
      return;
    }
    if (exception is Exception) {
      if (needLog) {
        logger.e('❌❌ Error: ${exception.toString()} ❌❌');
      }
      add(BaseErrorEvent(
        tag: tag,
        exception: KntException(
          KntIssue.undefined,
          additionalInfo: exception.toString(),
        ),
      ));
    }
  }

  /// Using whenever a event which fetch data from remote is added to bloc
  Future<void> checkNetworkConnection() async {
    final networkStatus = await Connectivity().checkConnectivity();
    if (networkStatus == ConnectivityResult.none) {
      throw KntException(KntIssue.noInternet);
    }
  }

  /// Wrap cancelable async process with connection status - [needCheckConnection]
  /// and error process if need - [needHandleError]
  Future? cancelableFuture(
    Future inner, {
    bool needCheckConnection = true,
    bool needHandleError = true,
  }) async {
    final preProcess =
        needCheckConnection ? () async => await checkNetworkConnection() : null;
    final process = processCancelableFuture(inner, preProcess: preProcess);
    return needHandleError
        ? process?.onError(
            (error, stackTrace) => handleError(error, tag: _currentEvent?.tag))
        : process;
  }

  /// Create an async process which can be cancelled
  Future? processCancelableFuture(
    Future inner, {
    Future? Function()? preProcess,
  }) async {
    await preProcess?.call();
    await _cancelableOperation?.cancel();
    _cancelableOperation = CancelableOperation.fromFuture(inner);
    return _cancelableOperation?.value;
  }
}

typedef CompareFunction<T> = int Function(T, T);

/// Base handles network required events & specific [DioError]
abstract class BaseApiBloc extends BaseBloc {
  BaseApiBloc({bool needLog = false}) : super(needLog: needLog);

  @override
  void handleError(exception, {String? tag}) {
    if (exception is DioError) {
      add(BaseErrorEvent(
        tag: tag,
        exception: KntException(
          exception.response?.statusCode?.convertedIssue,
          additionalInfo: exception.response?.statusMessage,
        ),
      ));
      return;
    }
    super.handleError(exception, tag: tag);
  }
}

/// Base handle paging events
abstract class BasePagingBloc<Q, D> extends BaseApiBloc {
  final List<D> _dataList = [];
  late int _currentOffset;
  late bool _loadedAllData;

  BasePagingBloc({bool needLog = false}) : super(needLog: needLog) {
    _init();
  }

  @override
  Stream<BaseState> handleEvent(BaseEvent event) async* {
    // Handle paging event
    if (event is BasePagingEvent<Q>) {
      // Check if all data is loaded
      final isFirstRequest = event.isFirstRequest;
      if (!isFirstRequest && _loadedAllData) {
        return;
      }
      // Check conditions
      if (event.needCheckConnection) {
        await checkNetworkConnection();
        // Add other conditions in the future
      }
      if (isFirstRequest) {
        _init();
      }

      // After that, start paging request
      final tag = event.tag;
      yield BaseLoadingState(tag: tag);
      final responsePaging = await cancelableFuture(
        pagingInProgress(BaseRequestPaging<Q>(
          queryData: event.queryData,
          limit: event.limit,
          offset: _currentOffset,
        )),
        // Already checked connection
        needCheckConnection: false,
        // Need rethrow here to spread out to outer layer
        needHandleError: false,
      );
      if (responsePaging is! BaseResponsePaging<D>) {
        throw KntException(KntIssue.parsing);
      }
      List<D> dataList = responsePaging.dataList;
      if (dataList.length < event.limit) {
        // Reached last page if result has less than limit
        _loadedAllData = true;
      }
      if (dataList.isEmpty && isFirstRequest) {
        // In case at first paging request we have empty data, we could stop
        yield BaseEmptyResultState(tag: tag);
        return;
      }
      // Update data list and offset
      _dataList.addAll(dataList.where((element) => element != null));
      if (responsePaging.currentOffset == defaultInvalidOffset) {
        _currentOffset += dataList.length;
      } else {
        _currentOffset = responsePaging.currentOffset;
      }
      // Sorting
      if (comparable != null) {
        _dataList.sort(comparable);
      }
      // Show state
      yield pagingStateInstance(_dataList, _currentOffset, tag: tag);
      return;
    }
    // Let parents handle the others
    yield* super.handleEvent(event);
  }

  @override
  Stream<BaseState> handleBaseErrorEvent(BaseErrorEvent event) async* {
    if (_dataList.isEmpty) {
      // Error occurs at the first query
      yield* super.handleBaseErrorEvent(event);
      return;
    }
    // Otherwise, ignore this error while loading more data
    yield pagingStateInstance(_dataList, _currentOffset, tag: event.tag);
  }

  /// Handle query process
  Future<BaseResponsePaging<D>> pagingInProgress(
      BaseRequestPaging<Q> pagingQuery);

  /// Create instance of paging result
  BasePagingResultState<D> pagingStateInstance(
    List<D> dataList,
    int offset, {
    String? tag,
  });

  /// Override this getter when we want to sort out result
  CompareFunction<D>? get comparable => null;

  void _init() {
    _dataList.clear();
    _currentOffset = defaultOffsetPaging;
    _loadedAllData = false;
  }
}
